﻿using System;
using System.Collections;
using System.Collections.Generic;
using System.Runtime.InteropServices;
using System.ServiceProcess;
using System.Reflection;
using System.IO;
using System.Threading;
using System.Net;
using System.Net.Sockets;
using System.Net.NetworkInformation;
using System.Xml;
using System.Text;
using System.Text.RegularExpressions;
using System.Configuration.Install;
using System.ComponentModel;
using System.Diagnostics;
using System.Linq;
using ARSoft.Tools.Net.Dns;

namespace it.lorenzomonti.vpndns
{
    internal class VpnDnsConfigException : Exception
    {
        public VpnDnsConfigException(string message)
            : base(message)
        {
        }
    }

    internal class VpnDnsConstants
    {
        public const string ServiceName = "VpnDnsSvc";
        public const string Service = "VPN remapping DNS proxy";
        public const string Log = "System";
    }

    internal class VpnDnsIPUtils
    {
        private static Regex arpa_re = new Regex(
            @"^(?!\.)(((?:^|\.)((25[0-5]|2[0-4][0-9]|1[0-9][0-9]|[1-9]?[0-9]))){1,4})\.in-addr.arpa$",
            RegexOptions.Compiled
        );

        private static Regex net_re = new Regex(
            @"^(25[0-5]|2[0-4][0-9]|1[0-9][0-9]|[1-9]?[0-9])\.(25[0-5]|2[0-4][0-9]|1[0-9][0-9]|[1-9]?[0-9])\.(25[0-5]|2[0-4][0-9]|1[0-9][0-9]|[1-9]?[0-9])\.(25[0-5]|2[0-4][0-9]|1[0-9][0-9]|[1-9]?[0-9])/(3[0-2]|[1-2]?[0-9])$",
            RegexOptions.Compiled
        );

        public static IPAddress ParseArpaName(string name, out int count)
        {
            count = 0;

            Match match = arpa_re.Match(name);

            if (match.Success && match.Groups.Count >= 2)
            {
                string[] nums = Regex.Split(match.Groups[1].Value, @"\.");
                byte[] data = new byte[4];

                for (int i = nums.Length; i > 0; i--)
                    data[nums.Length - i] = byte.Parse(nums[i - 1]);

                count = nums.Length;

                return new IPAddress(data);
            }

            return null;
        }

        public static bool ParseNet(string value, out IPAddress ip, out ushort maskbits)
        {
            ip = null;
            maskbits = 0;

            Match match = net_re.Match(value);

            if (match.Success && match.Groups.Count >= 6)
            {
                byte[] data = new byte[4];
                data[0] = byte.Parse(match.Groups[1].Value);
                data[1] = byte.Parse(match.Groups[2].Value);
                data[2] = byte.Parse(match.Groups[3].Value);
                data[3] = byte.Parse(match.Groups[4].Value);

                ip = new IPAddress(data);
                maskbits = ushort.Parse(match.Groups[5].Value);

                return true;
            }

            return false;
        }

        public static string BuildArpaName(IPAddress ip, int count)
        {
            byte[] data = ip.GetAddressBytes();

            StringBuilder sb = new StringBuilder();

            for (int i = count - 1; i >= 0; i--)
            {
                sb.Append((int)data[i]);
                sb.Append('.');
            }
            
            sb.Append("in-addr.arpa");

            return sb.ToString();
        }

        public static List<IPAddress> GetSystemDnsServers()
        {
            List<IPAddress> list = new List<IPAddress>();

            foreach (NetworkInterface iface in NetworkInterface.GetAllNetworkInterfaces())
            {
                if (!iface.Supports(NetworkInterfaceComponent.IPv4) ||
                    iface.OperationalStatus == OperationalStatus.Down)
                {
                    continue;
                }

                foreach (IPAddress addr in iface.GetIPProperties().DnsAddresses)
                {
                    if (addr.AddressFamily != AddressFamily.InterNetwork ||
                        IPAddress.IsLoopback(addr))
                    {
                        continue;
                    }

                    list.Add(addr);
                }
            }

            return list;
        }
    }

    internal class VpnDnsRemapData
    {
        private byte[] m_from;
        private byte[] m_to;
        private byte[] m_netmask = new byte[4];
        private byte[] m_hostmask = new byte[4];
        private ushort m_maskbits;

        public IPAddress From
        {
            get
            {
                return new IPAddress(m_from);
            }
        }

        public IPAddress To
        {
            get
            {
                return new IPAddress(m_to);
            }
        }

        public ushort Maskbits
        {
            get
            {
                return m_maskbits;
            }
        }

        public VpnDnsRemapData(IPAddress from, IPAddress to, ushort maskbits)
        {
            m_maskbits = maskbits;

            uint mask = (uint)(((ulong)1 << maskbits) - 1);
            m_netmask[0] = (byte)(((mask) & 0xff));
            m_netmask[1] = (byte)(((mask >> 8) & 0xff));
            m_netmask[2] = (byte)(((mask >> 16) & 0xff));
            m_netmask[3] = (byte)(((mask >> 24) & 0xff));

            m_hostmask[0] = (byte)(((~mask) & 0xff));
            m_hostmask[1] = (byte)(((~mask >> 8) & 0xff));
            m_hostmask[2] = (byte)(((~mask >> 16) & 0xff));
            m_hostmask[3] = (byte)(((~mask >> 24) & 0xff));

            m_from = from.GetAddressBytes();
            m_to = to.GetAddressBytes();

            m_from[0] &= m_netmask[0];
            m_from[1] &= m_netmask[1];
            m_from[2] &= m_netmask[2];
            m_from[3] &= m_netmask[3];

            m_to[0] &= m_netmask[0];
            m_to[1] &= m_netmask[1];
            m_to[2] &= m_netmask[2];
            m_to[3] &= m_netmask[3];
        }

        public IPAddress RemapIP(IPAddress ip, bool reverse = false)
        {
            byte[] src_net = ip.GetAddressBytes();
            byte[] src_ip = ip.GetAddressBytes();

            src_net[0] &= m_netmask[0];
            src_net[1] &= m_netmask[1];
            src_net[2] &= m_netmask[2];
            src_net[3] &= m_netmask[3];

            byte[] chk_net = reverse
                ? m_to
                : m_from;

            byte[] dst_net = reverse
                ? m_from
                : m_to;

            if (src_net[0] != chk_net[0] ||
                src_net[1] != chk_net[1] ||
                src_net[2] != chk_net[2] ||
                src_net[3] != chk_net[3])
            {
                return ip;
            }

            src_ip[0] &= m_hostmask[0];
            src_ip[1] &= m_hostmask[1];
            src_ip[2] &= m_hostmask[2];
            src_ip[3] &= m_hostmask[3];

            src_ip[0] |= dst_net[0];
            src_ip[1] |= dst_net[1];
            src_ip[2] |= dst_net[2];
            src_ip[3] |= dst_net[3];

            return new IPAddress(src_ip);
        }

        public bool IsSameSubnet(IPAddress ip, bool reverse = false)
        {
            byte[] src_net = ip.GetAddressBytes();

            src_net[0] &= m_netmask[0];
            src_net[1] &= m_netmask[1];
            src_net[2] &= m_netmask[2];
            src_net[3] &= m_netmask[3];

            byte[] chk_net = reverse
                ? m_from
                : m_to;

            return (src_net[0] == chk_net[0] &&
                src_net[1] == chk_net[1] &&
                src_net[2] == chk_net[2] &&
                src_net[3] == chk_net[3]);
        }
    }

    internal class VpnDnsZone
    {
        private bool m_active = false;
        private VpnDnsRemapData m_remap;
        private Regex m_suffix;
        private string m_name;
        private string m_sentinel;
        private List<string> m_suffixes;
        private DnsClient m_client;
        List<IPAddress> m_dnsServers;
        private static Regex dnscheck = new Regex(
            @"^([a-z0-9]+(-[a-z0-9]+)*\.)+[a-z]{2,}$",
            RegexOptions.Compiled | RegexOptions.ExplicitCapture
        );

        public bool Active
        {
            get
            {
                return m_active;
            }
        }

        public void OnFileCreated(string name)
        {
            if (!m_active && name.Equals(m_sentinel, StringComparison.InvariantCultureIgnoreCase))
            {
                m_active = true;
                VpnDnsService.LogMessage(string.Format("Zone \"{0}\" has been activated.", m_name));
            }
        }

        public void OnFileDeleted(string name)
        {
            if (m_active && name.Equals(m_sentinel, StringComparison.InvariantCultureIgnoreCase))
            {
                m_active = false;
                VpnDnsService.LogMessage(string.Format("Zone {0} deactivated.", m_name));
            }
        }

        private bool IsValidDomainName(string suffix)
        {
            return dnscheck.IsMatch(suffix);
        }

        public VpnDnsZone(string name, string suffixes, string sentinel, string sentineldir,
            VpnDnsRemapData remap, List<IPAddress> dnsServers, int queryTimeout)
        {
            m_name = name;
            m_sentinel = sentinel;
            m_suffixes = new List<string>();

            StringBuilder sb = new StringBuilder();

            foreach (string str in Regex.Split(suffixes, @"\|"))
            {
                string suffix = str.Trim(' ', '\t', '.').ToLowerInvariant();

                if (!IsValidDomainName(suffix))
                    throw new VpnDnsConfigException("invalid suffix " + str);

                m_suffixes.Add(suffix);

                if (sb.Length > 0)
                    sb.Append('|');
                sb.Append(@"(^|\.)");
                sb.Append(suffix.Replace(".", @"\."));
                sb.Append(@"(\.)?$");
            }

            m_suffix = new Regex(sb.ToString(),
                RegexOptions.Compiled | RegexOptions.IgnoreCase | RegexOptions.ExplicitCapture);
            m_remap = remap;
            m_dnsServers = dnsServers;
            m_client = new DnsClient(dnsServers, queryTimeout);

            if (sentineldir.Length > 0 &&
                m_sentinel.Length > 0 &&
                File.Exists(sentineldir + @"\" + m_sentinel))
            {
                m_active = true;
            }
        }

        public bool HandlesName(string name)
        {
            return m_suffix.IsMatch(name);
        }

        public bool HandlesIP(IPAddress ip, bool reverse = false)
        {
            return m_remap.IsSameSubnet(ip, reverse);
        }

        public IPAddress RemapIP(IPAddress ip, bool reverse = false)
        {
            return m_remap.RemapIP(ip, reverse);
        }

        private void AddRemappedRecord(DnsRecordBase record, List<DnsRecordBase> answers)
        {
            DnsRecordBase newrec = record;

            if (record.RecordType == RecordType.A && HandlesName(record.Name))
            {
                ARecord arecord = record as ARecord;
                newrec = new ARecord(arecord.Name, arecord.TimeToLive, RemapIP(arecord.Address));
            }
            else if (record.RecordType == RecordType.Ptr ||
                record.RecordType == RecordType.Ns ||
                record.RecordType == RecordType.Soa)
            {
                int count;
                IPAddress ip = VpnDnsIPUtils.ParseArpaName(record.Name, out count);

                if (ip != null && HandlesIP(ip, true))
                {
                    string name = VpnDnsIPUtils.BuildArpaName(RemapIP(ip), count);

                    switch (record.RecordType)
                    {
                        case RecordType.Ptr:
                            PtrRecord ptrrec = record as PtrRecord;
                            newrec = new PtrRecord(name, record.TimeToLive, ptrrec.PointerDomainName);
                            break;

                        case RecordType.Ns:
                            NsRecord nsrec = record as NsRecord;
                            newrec = new NsRecord(name, record.TimeToLive, nsrec.NameServer);
                            break;

                        case RecordType.Soa:
                            SoaRecord soarec = record as SoaRecord;
                            newrec = new SoaRecord(name, record.TimeToLive, soarec.MasterName, soarec.ResponsibleName,
                                soarec.SerialNumber, soarec.RefreshInterval, soarec.RetryInterval, soarec.ExpireInterval,
                                soarec.NegativeCachingTTL);
                            break;
                    }
                }
            }

            answers.Add(newrec);
        }

        public DnsMessage Resolve(DnsQuestion question, DnsMessage query, string name = null)
        {
            DnsMessage answer = m_client.Resolve((name == null) ? question.Name : name,
                question.RecordType, question.RecordClass);

            if (answer != null)
            {
                foreach (DnsRecordBase record in (answer.AnswerRecords))
                {
                    AddRemappedRecord(record, query.AnswerRecords);
                }

                foreach (DnsRecordBase record in (answer.AdditionalRecords))
                {
                    AddRemappedRecord(record, query.AdditionalRecords);
                }

                foreach (DnsRecordBase record in (answer.AuthorityRecords))
                {
                    AddRemappedRecord(record, query.AuthorityRecords);
                }

                query.ReturnCode = ReturnCode.NoError;
                return query;
            }

            query.ReturnCode = ReturnCode.ServerFailure;
            return query;
        }

        public string Description
        {
            get
            {
                StringBuilder sb = new StringBuilder();

                sb.AppendLine(String.Format("name = {0}", m_name));
                sb.AppendLine(String.Format("active = {0}", m_active));
                sb.Append("suffixes = ");
                foreach (string suffix in m_suffixes)
                {
                    sb.Append(suffix);
                    sb.Append(' ');
                }
                sb.AppendLine();
                sb.Append("using DNS server(s) = ");
                foreach (IPAddress ip in m_dnsServers)
                {
                    sb.Append(ip.ToString());
                    sb.Append(' ');
                }
                sb.AppendLine();
                sb.AppendLine(String.Format("remap = from {0}/{2} to {1}/{2}",
                    m_remap.From, m_remap.To, m_remap.Maskbits));

                return sb.ToString();
            }
        }
    }

    internal class VpnDnsCache
    {
        internal class CacheEntry
        {
            private DateTime m_expiredate;
            private DnsMessage m_message;

            public bool IsValid
            {
                get
                {
                    return m_expiredate > DateTime.Now;
                }
            }

            public DnsMessage Message
            {
                get
                {
                    return m_message;
                }
            }

            public CacheEntry(DnsMessage message)
            {
                m_message = message;
                m_expiredate = DateTime.Now;
                int minttl = -1;

                foreach (List<DnsRecordBase> list in new List<DnsRecordBase>[] {
                    message.AnswerRecords, message.AdditionalRecords, message.AuthorityRecords })
                {
                    foreach (DnsRecordBase rec in list)
                    {
                        if (rec.TimeToLive < 0)
                            return;
                        if (minttl < 0 || rec.TimeToLive < minttl)
                            minttl = rec.TimeToLive;
                    }
                }

                if (minttl > 0)
                    m_expiredate = m_expiredate.AddSeconds(minttl);
            }
        }

        private ReaderWriterLockSlim m_cacheLock = new ReaderWriterLockSlim();
        private Dictionary<string, CacheEntry> m_cache = new Dictionary<string, CacheEntry>();
        private Thread m_cleanup = null;

        private void CleanupProc()
        {
            while (true)
            {
                if (VpnDnsService.ShutdownEvent.WaitOne(120000))
                    return;

                m_cacheLock.EnterWriteLock();
                try
                {
                    while (true)
                    {
                        KeyValuePair<string, CacheEntry> pair = m_cache.FirstOrDefault(element => !element.Value.IsValid);
                        if (pair.Key == null)
                            break;
                        m_cache.Remove(pair.Key);
                    }
                }
                finally
                {
                    m_cacheLock.ExitWriteLock();
                }
            }
        }

        public VpnDnsCache()
        {
            m_cleanup = new Thread(new ThreadStart(CleanupProc));
            m_cleanup.Start();
        }

        ~VpnDnsCache()
        {
            m_cleanup.Join();
        }

        public bool GetCachedMessage(string key, out DnsMessage message)
        {
            bool result = false;
            CacheEntry cached_result = null;
            message = null;
            
            m_cacheLock.EnterReadLock();
            try
            {
                result = m_cache.TryGetValue(key, out cached_result) && cached_result.IsValid;
                if (result)
                    message = cached_result.Message;
            }
            finally
            {
                m_cacheLock.ExitReadLock();
            }

            return result;
        }

        public void SetCachedMessage(string key, DnsMessage message)
        {
            m_cacheLock.EnterWriteLock();
            try
            {
                m_cache.Add(key, new CacheEntry(message));
            }
            finally
            {
                m_cacheLock.ExitWriteLock();
            }
        }
    }

    internal class VpnDnsConfig
    {
        private List<IPAddress> m_dnsservers;
        private VpnDnsZone[] m_zones;
        private int m_queryTimeout = 5000;
        private FileSystemWatcher m_watcher = null;
        private string m_sentineldir = "";
        private VpnDnsCache m_cache = new VpnDnsCache();

        public List<IPAddress> SystemDnsServers
        {
            get
            {
                return m_dnsservers;
            }
        }

        public int QueryTimeout
        {
            get
            {
                return m_queryTimeout;
            }
        }

        public string SentinelDir
        {
            get
            {
                return m_sentineldir;
            }
        }

        public VpnDnsConfig()
            : this("config.xml")
        {
        }

        public VpnDnsConfig(string filename)
        {
            XmlDocument xml = new XmlDocument();
            xml.Load(filename);

            ReadDnsServers(xml);
            ReadVpnZones(xml);
        }

        protected void ReadDnsServers(XmlDocument xml)
        {
            XmlNode serversnode = xml.SelectSingleNode(@"/vpndns/dnsservers");

            if (serversnode != null)
            {
                XmlNodeList list = serversnode.SelectNodes(@"dnsserver/@address");

                if (list == null || list.Count == 0)
                {
                    m_dnsservers = VpnDnsIPUtils.GetSystemDnsServers();
                }
                else
                {
                    m_dnsservers = new List<IPAddress>();
                    foreach (XmlNode node in list)
                    {
                        IPAddress ip;
                        if (IPAddress.TryParse(node.Value, out ip))
                            m_dnsservers.Add(ip);
                    }
                }

                XmlNode attr = serversnode.Attributes["timeout"];
                if (attr != null)
                    m_queryTimeout = int.Parse(attr.Value);
            }

            if (m_dnsservers == null || m_dnsservers.Count == 0)
            {
                VpnDnsService.LogMessage("Warning: no valid DNS servers found. Trying to use the last successfully used one(s).", EventLogEntryType.Warning);
                m_dnsservers = ReadPreviosDnsServers();
            }
            else
                WriteDnsServers(m_dnsservers);

            if (m_dnsservers == null || m_dnsservers.Count == 0)
                throw new VpnDnsConfigException("no DNS servers");

            foreach (IPAddress ip in m_dnsservers)
            {
                VpnDnsService.LogMessage("Using DNS server(s): " + ip.ToString());
            }
        }

        private List<IPAddress> ReadPreviosDnsServers()
        {
            try
            {
                List<IPAddress> list = new List<IPAddress>();
                string file = VpnDnsService.AssemblyDirectory + @"\dnsstate.xml";

                XmlDocument xml = new XmlDocument();
                xml.Load(file);

                foreach (XmlNode node in xml.SelectNodes(@"/dnsservers/dnsserver/@address"))
                {
                    IPAddress ip;
                    if (IPAddress.TryParse(node.Value, out ip))
                        list.Add(ip);
                }

                return list;
            }
            catch
            {
                return null;
            }
        }

        private void WriteDnsServers(List<IPAddress> list)
        {
            try
            {
                XmlDocument xml = new XmlDocument();
                xml.AppendChild(xml.CreateXmlDeclaration("1.0", "UTF-8", ""));
                XmlNode root = xml.CreateElement("dnsservers");

                foreach (IPAddress ip in list)
                {
                    XmlNode dns = xml.CreateElement("dnsserver");
                    XmlAttribute attr = xml.CreateAttribute("address");
                    attr.Value = ip.ToString();
                    dns.Attributes.Append(attr);
                    root.AppendChild(dns);
                }

                xml.AppendChild(root);

                string file = VpnDnsService.AssemblyDirectory + @"\dnsstate.xml";

                xml.Save(file);
            }
            catch (Exception e)
            {
                VpnDnsService.LogMessage(e.Message + Environment.NewLine + e.StackTrace, EventLogEntryType.Error);
            }
        }

        private void OnFileDeleted(object source, FileSystemEventArgs e)
        {
            if (VpnDnsService.RunningInConsole)
                Console.Out.WriteLine(e.Name + " deleted");
            foreach (VpnDnsZone zone in m_zones.Where(zone => zone.Active))
                zone.OnFileDeleted(e.Name);
        }

        private void OnFileCreated(object source, FileSystemEventArgs e)
        {
            if (VpnDnsService.RunningInConsole)
                Console.Out.WriteLine(e.Name + " created");
            foreach (VpnDnsZone zone in m_zones.Where(zone => !zone.Active))
                zone.OnFileCreated(e.Name);
        }

        private void OnFileRenamed(object source, RenamedEventArgs e)
        {
            if (VpnDnsService.RunningInConsole)
                Console.WriteLine("{0} renamed to {1}", e.OldName, e.Name);
            foreach (VpnDnsZone zone in m_zones)
            {
                zone.OnFileDeleted(e.OldName);
                zone.OnFileCreated(e.Name);
            }
        }


        protected void ReadVpnZones(XmlDocument xml)
        {
            XmlNode attr;

            attr = xml.SelectSingleNode(@"/vpndns/zones/sentineldir/@path");
            if (attr != null)
            {
                m_sentineldir = attr.Value;
                attr = xml.SelectSingleNode(@"/vpndns/zones/sentineldir/@filter");
                if (attr != null)
                {
                    string filter = attr.Value;
                    m_watcher = new FileSystemWatcher(m_sentineldir, filter);
                }
                else
                {
                    m_watcher = new FileSystemWatcher(m_sentineldir);
                }

                m_watcher.NotifyFilter = NotifyFilters.FileName;
                m_watcher.Deleted += new FileSystemEventHandler(OnFileDeleted);
                m_watcher.Created += new FileSystemEventHandler(OnFileCreated);
                m_watcher.Renamed += new RenamedEventHandler(OnFileRenamed);
            }

            ArrayList al = new ArrayList();

            foreach (XmlNode zonenode in xml.SelectNodes(@"/vpndns/zones/zone"))
            {
                XmlNode child;
                string suffixes, name, sentinel;

                attr = zonenode.Attributes["name"];
                if (attr != null)
                    name = attr.Value;
                else
                    name = "<no name>";

                attr = zonenode.Attributes["sentinel"];
                if (attr != null)
                    sentinel = attr.Value;
                else
                    sentinel = "";

                attr = zonenode.Attributes["suffixes"];
                if (attr == null || (suffixes = attr.Value.Trim()).Length == 0)
                    throw new VpnDnsConfigException("no suffixes defined for zone " + name);

                VpnDnsRemapData remap = null;
                child = zonenode.SelectSingleNode(@"remap");

                if (child != null)
                {
                    XmlNode from, to;
                    from = child.Attributes["from"];
                    to = child.Attributes["to"];
                    if (from != null && to != null)
                    {
                        remap = GetRemap(from.Value, to.Value);
                    }
                }

                if (remap == null)
                    throw new VpnDnsConfigException("invalid remap for zone " + name);

                List<IPAddress> dnsservers = new List<IPAddress>();
                int queryTimeout = 5000;

                XmlNode serversnode = zonenode.SelectSingleNode(@"dnsservers");

                if (serversnode != null)
                {
                    foreach (XmlNode node in serversnode.SelectNodes(@"dnsserver/@address"))
                    {
                        IPAddress ip;
                        if (IPAddress.TryParse(node.Value, out ip))
                            dnsservers.Add(remap.RemapIP(ip));
                    }

                    attr = serversnode.Attributes["timeout"];
                    if (attr != null)
                        queryTimeout = int.Parse(attr.Value);
                }

                if (dnsservers.Count == 0)
                    throw new VpnDnsConfigException("invalid DNS servers for zone " + name);

                VpnDnsZone vpnzone = new VpnDnsZone(name, suffixes, sentinel, m_sentineldir,
                    remap, dnsservers, queryTimeout);
                al.Add(vpnzone);

                VpnDnsService.LogMessage("starting a new VPN zone" + Environment.NewLine + vpnzone.Description);
            }

            m_zones = (VpnDnsZone[])al.ToArray(typeof(VpnDnsZone));

            if (m_watcher != null)
                m_watcher.EnableRaisingEvents = true;
        }

        protected VpnDnsRemapData GetRemap(string from, string to)
        {
            IPAddress ip_from = null;
            IPAddress ip_to = null;
            ushort maskbits_from, maskbits_to;

            if (VpnDnsIPUtils.ParseNet(from, out ip_from, out maskbits_from) &&
                VpnDnsIPUtils.ParseNet(to, out ip_to, out maskbits_to) &&
                maskbits_from == maskbits_to)
            {
                return new VpnDnsRemapData(ip_from, ip_to, maskbits_from);
            }

            return null;
        }

        public DnsMessage Resolve(DnsQuestion question, DnsMessage query)
        {
            DnsMessage cached_result = null;
            string key = question.ToString();

            if (m_cache.GetCachedMessage(key, out cached_result))
            {
                cached_result.TransactionID = query.TransactionID;
#if DEBUG
                if (VpnDnsService.RunningInConsole)
                    Console.Out.WriteLine(key + " : cache hit!");
#endif
            }
            else
            {
                foreach (VpnDnsZone zone in m_zones.Where(zone => zone.Active))
                {
                    if (zone.HandlesName(question.Name))
                    {
                        cached_result = zone.Resolve(question, query);
                        break;
                    }
                    else
                    {
                        int count;
                        IPAddress ip = VpnDnsIPUtils.ParseArpaName(question.Name, out count);

                        if (ip != null && zone.HandlesIP(ip))
                        {
                            string name = VpnDnsIPUtils.BuildArpaName(zone.RemapIP(ip, true), count);
                            cached_result = zone.Resolve(question, query, name);
                            break;
                        }
                    }
                }

                if (cached_result != null &&
                    cached_result.ReturnCode == ReturnCode.NoError &&
                    cached_result.AnswerRecords.Count != 0)
                {
                    m_cache.SetCachedMessage(key, cached_result);
                }
            }

            return cached_result;
        }
    }

    class VpnDnsService : ServiceBase
    {
        private static VpnDnsConfig m_conf = null;
        private static DnsClient m_systemdns = null;
        private static DnsServer m_server = null;
        private static VpnDnsService m_service = null;
        private static ManualResetEvent m_shutdownEvent = new ManualResetEvent(false);
        private static bool m_bConsole = false;

        static public string AssemblyDirectory
        {
            get
            {
                string codeBase = Assembly.GetExecutingAssembly().CodeBase;
                UriBuilder uri = new UriBuilder(codeBase);
                string path = Uri.UnescapeDataString(uri.Path);
                return Path.GetDirectoryName(path);
            }
        }

        public static ManualResetEvent ShutdownEvent
        {
            get
            {
                return m_shutdownEvent;
            }
        }

        public static void LogMessage(string message, EventLogEntryType type = EventLogEntryType.Information)
        {
            if (RunningInConsole)
            {
                if (type == EventLogEntryType.Information)
                    Console.Out.WriteLine(message);
                else
                    Console.Error.WriteLine(message);
            }
            else
            {
                EventLog.WriteEntry(VpnDnsConstants.Service, message, type);
            }
        }

        public static bool RunningInConsole
        {
            get
            {
                return m_bConsole;
            }
        }

        public VpnDnsService()
        {
        }

        protected override void OnStart(string[] args)
        {
            try
            {
                if (RunningInConsole)
                    Console.Out.WriteLine("starting server");

#if DEBUG
                m_conf = new VpnDnsConfig(@"..\Release\config.xml");
#else
                m_conf = new VpnDnsConfig(VpnDnsService.AssemblyDirectory + @"\config.xml");
#endif
                m_systemdns = new DnsClient(m_conf.SystemDnsServers, m_conf.QueryTimeout);

                m_server = new DnsServer(IPAddress.Loopback, 10, 10, ProcessQuery);
                m_server.Start();

                base.OnStart(args);
            }
            catch (Exception e)
            {
                LogMessage(e.Message + Environment.NewLine + e.StackTrace, EventLogEntryType.Error);
                Stop();
            }
        }

        protected override void OnStop()
        {
            if (RunningInConsole)
                Console.Out.WriteLine("stopping server");

            if (m_server != null)
                m_server.Stop();
            m_shutdownEvent.Set();

            base.OnStop();
        }

        static DnsMessageBase ProcessQuery(DnsMessageBase message, IPAddress clientAddress, ProtocolType protocol)
        {
            message.IsQuery = false;

            DnsMessage query = message as DnsMessage;

            if ((query != null) && (query.Questions.Count == 1))
            {
                DnsQuestion question = query.Questions[0];

                DnsMessageBase ret = m_conf.Resolve(question, query);

                if (ret != null)
                    return ret;

                DnsMessage answer = m_systemdns.Resolve(question.Name, question.RecordType, question.RecordClass);

                if (answer != null)
                {
                    foreach (DnsRecordBase record in (answer.AnswerRecords))
                    {
                        query.AnswerRecords.Add(record);
                    }

                    foreach (DnsRecordBase record in (answer.AdditionalRecords))
                    {
                        query.AdditionalRecords.Add(record);
                    }

                    foreach (DnsRecordBase record in (answer.AuthorityRecords))
                    {
                        query.AuthorityRecords.Add(record);
                    }

                    query.ReturnCode = ReturnCode.NoError;
                    return query;
                }
            }

            message.ReturnCode = ReturnCode.ServerFailure;
            return message;
        }
        
        private static void BreakHandler(object sender, ConsoleCancelEventArgs args)
        {
            m_service.Stop();
        }

        static void Main(string[] args)
        {
            try
            {
                if (args.Length > 0 && args[0].Equals("--console", StringComparison.InvariantCultureIgnoreCase))
                    m_bConsole = true;
                else
                    m_bConsole = false;

                if (!RunningInConsole && !EventLog.SourceExists(VpnDnsConstants.Service))
                    EventLog.CreateEventSource(VpnDnsConstants.Service, VpnDnsConstants.Log);

                m_service = new VpnDnsService();

                if (RunningInConsole)
                {
                    Console.CancelKeyPress += new ConsoleCancelEventHandler(BreakHandler);
                    m_service.OnStart(args);
                    m_shutdownEvent.WaitOne();
                }
                else
                {
                    ServiceBase.Run(m_service);
                }
            }
            catch (Exception e)
            {
                LogMessage(e.Message + Environment.NewLine + e.StackTrace, EventLogEntryType.Error);
            }
        }
    }

    [RunInstaller(true)]
    public class VpnDnsInstaller : Installer
    {
        public VpnDnsInstaller()
        {
            ServiceProcessInstaller process = new ServiceProcessInstaller();

            process.Account = ServiceAccount.LocalSystem;

            ServiceInstaller service = new ServiceInstaller();

            service.StartType = ServiceStartMode.Manual;
            service.ServiceName = VpnDnsConstants.ServiceName;
            service.DisplayName = VpnDnsConstants.Service;

            Installers.Add(process);
            Installers.Add(service);
        }
    }
}
